from stackboost.utils.data_manipulation import to_array
import numpy as np
from numba import float32, boolean, int32, char
from numba.core.types.npytypes import Array
from numba.core.types.misc import UnicodeType
from numba.experimental import jitclass
import torch
import pandas as pd
import numba
from stackboost.utils.errors import NotFittedError

SUPPORTED_ARRAYS = [np.ndarray, torch.tensor, pd.Series, pd.DataFrame, list]


class StackedEncoder:
    """
    Categorical feature incoding model
    """

    def __init__(self, strategy: str = "mean", regression: bool = True):
        self.strategy = strategy
        if self.strategy != None:
            self.strategies = {"mean": np.mean,
                               "median": np.median,
                               "sum": np.sum}

            self.strategy = self.strategy.lower()
            if self.strategy not in list(self.strategies.keys()):
                raise ValueError("Given strategy type", self.strategy,
                                 "allowed," ', '.join(list(self.strategies.keys())))
        self.regression = regression
        self.fitted = False

    def fit(self, X, y, cat_features=None) -> None:
        self.fitted = False
        X = to_array(X)
        y = to_array(y)

        self.n_features = X.shape[1]

        self.cat_features = cat_features
        if type(self.cat_features) not in SUPPORTED_ARRAYS:
            self.cat_features = np.array([i for i in range(self.n_features)])

        self.__target(X, y)

        self.fitted = True

    def __target(self, X: np.ndarray, y: np.ndarray) -> None:
        if self.regression:
            self.mappers = np.array([])
            for i in range(len(self.cat_features)):
                counts = dict()
                feature = X[:, self.cat_features[i]]
                uniques = np.unique(feature)

                for unique in uniques:
                    counts[unique] = self.strategies.get(self.strategy)(y[feature == unique])

                self.mappers = np.append(self.mappers, counts)
        else:
            self.mappers = np.array([])
            for i in range(len(self.cat_features)):
                counts = dict()
                feature = X[:, self.cat_features[i]]
                uniques = np.unique(feature)

                for unique in uniques:
                    counts[unique] = y[feature == unique].mean()

                self.mappers = np.append(self.mappers, counts)

    def transform(self, X: np.ndarray) -> np.ndarray:
        if not self.fitted:
            raise NotFittedError()

        X_array = to_array(X)
        n_features_test = X_array.shape[1]
        if self.n_features != n_features_test:
            raise ValueError(
                f"Shape of fitting array ({self.n_features} columns) and transforming array ({n_features_test} columns) do not match")

        for i in range(len(self.cat_features)):
            X_array[:, self.cat_features[i]] = np.vectorize(self.mappers[i].get)(X_array[:, self.cat_features[i]])

        return X_array

    def fit_transform(self, X, y=None, cat_features=None) -> np.ndarray:
        self.fit(X, y, cat_features)
        return self.transform(X)



from stackboost.models.tree import SimilarityTreeRegressor, DispersionTreeRegressor, ErrorTreeRegressor, StackedTreeClassifier
from stackboost.models.gradient_boosting import StackedGradientBoostingRegressor
from stackboost.loss_functions import MSE, SquareLoss, CrossEntropy

__author__ = 'Lev Novitskiy > levnovitskiy@gmail.com > https://github.com/leffff/ > https://www.kaggle.com/levnovitskiy'
__license__ = 'MIT'



from sklearn.datasets import load_boston
from sklearn.linear_model import LinearRegression
from stackboost.loss_functions import MSE, SquareLoss
from stackboost.utils.data_operation import mean_squared_error
import numpy as np
from catboost import CatBoostRegressor
from stackboost.loss_functions import SquareLoss
from stackboost import ErrorTreeRegressor, DispersionTreeRegressor, SimilarityTreeRegressor
from stackboost import StackedGradientBoostingRegressor

X, y = load_boston(return_X_y=True)
from sklearn.model_selection import KFold

kf = KFold(n_splits=5, shuffle=True, random_state=42)
kf.get_n_splits(X)

ws = []
gs = []
ds = []
dis = []
xs = []

c = 0
for train_index, test_index in kf.split(X):
    X_train, X_test = X[train_index], X[test_index]
    y_train, y_test = y[train_index], y[test_index]

    dtc = SimilarityTreeRegressor(n_quantiles=33, l2_leaf_reg=0.1, max_depth=7)
    dtc.fit(X_train, y_train)
    preds = dtc.predict(X_test)
    score = mean_squared_error(y_test, preds)
    print(score)
    ws.append(score)

    # SGBR = StackedGradientBoostingRegressor(estimator=SimilarityTreeRegressor(l2_leaf_reg=0, n_quantiles=40, loss_function=MSE()), n_estimators=100, learning_rate=0.04, loss_function=MSE())
    # SGBR.fit(X_train, y_train, eval_set=(X_test, y_test), cat_features=np.array([3, 8]), use_best_model=True)
    # preds = SGBR.predict(X_test)
    # score = mean_squared_error(y_test, preds)
    # print(score)
    # gs.append(score)
    # SGBR.plot()



    print("<------------>")


print("Wave", sum(ws) / len(ws))
print("GBR", sum(gs) / len(gs))
# print("DTR", sum(ds) / len(ds))
# print("ADA", sum(ads) / len(ads))
# print("DIS", sum(dis) / len(dis))
# print("XGBR", sum(xs) / len(xs))




import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="stackboost",
    version="0.1.0",
    author="leffff",
    author_email="levnovitskiy@gmail.com",
    description="Stacked decision trees algorithm",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/leffff/stackboost",
    license='MIT',
    packages=setuptools.find_packages(),
    install_requires=["numpy", "pandas", "torch", "plotly", "numba"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)


import numpy as np
from sklearn.metrics import accuracy_score
from numba.experimental import jitclass


class Loss(object):
    def loss(self, y_true, y_pred):
        return NotImplementedError()

    def gradient(self, y, y_pred):
        raise NotImplementedError()

    def acc(self, y, y_pred):
        return 0


@jitclass()
class SquareLoss(Loss):
    def __init__(self): pass

    def loss(self, y, y_pred):
        return np.sum(0.5 * np.power((y - y_pred), 2))

    def gradient(self, y, y_pred):
        return -(y - y_pred)

    def hess(self, y, y_pred):
        return np.array([1 for i in range(len(y))])


@jitclass()
class CrossEntropy(Loss):
    def __init__(self): pass

    def loss(self, y, p):
        p = np.clip(p, 1e-15, 1 - 1e-15)
        return - y * np.log(p) - (1 - y) * np.log(1 - p)

    def acc(self, y, p):
        return accuracy_score(np.argmax(y, axis=1), np.argmax(p, axis=1))

    def gradient(self, y, p):
        p = np.clip(p, 1e-15, 1 - 1e-15)
        return - (y / p) + (1 - y) / (1 - p)


@jitclass()
class MSE(Loss):
    def __init__(self): pass

    def loss(self, y, y_pred):
        return np.mean((y - y_pred) ** 2)

    def gradient(self, y, y_pred):
        return -(2 / len(y) * (y - y_pred))